@@ -1,15 +1,17 @@
-/* $Id: unshield.c,v 1.20 2005/07/08 14:08:37 twogood Exp $ */
+/* $Id: unshield.c 2995 2007-08-05 07:47:46Z twogood $ */
+#ifdef __linux__
 #define _BSD_SOURCE 1
 #define _POSIX_C_SOURCE 2
-#include "../lib/libunshield.h"
+#endif
+#include <sys/types.h>
+#include <sys/stat.h>
 #include <ctype.h>
 #include <locale.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
-#include <sys/types.h>
-#include <sys/stat.h>
 #include <unistd.h>
+#include "../lib/libunshield.h"
 #ifdef HAVE_CONFIG_H
 #include "../lib/unshield_config.h"
 #endif
@@ -36,6 +38,13 @@ typedef enum
   ACTION_TEST
 } ACTION;
 
+typedef enum
+{
+  FORMAT_NEW,
+  FORMAT_OLD,
+  FORMAT_RAW
+} FORMAT;
+
 #define DEFAULT_OUTPUT_DIRECTORY  "."
 
 static const char* output_directory   = DEFAULT_OUTPUT_DIRECTORY;
@@ -48,6 +57,7 @@ static ACTION action                  = ACTION_EXTRACT;
 static OVERWRITE overwrite            = OVERWRITE_ASK;
 static int log_level                  = UNSHIELD_LOG_LEVEL_LOWEST;
 static int exit_status                = 0;
+static FORMAT format                  = FORMAT_NEW;
 
 static bool make_sure_directory_exists(const char* directory)/*{{{*/
 {
@@ -98,7 +108,7 @@ static void show_usage(const char* name)
   fprintf(stderr,
       "Syntax:\n"
       "\n"
-      "\t%s [-c COMPONENT] [-d DIRECTORY] [-D LEVEL] [-g GROUP] [-G] [-h] [-l] [-V] c|l|x CABFILE\n"
+      "\t%s [-c COMPONENT] [-d DIRECTORY] [-D LEVEL] [-g GROUP] [-GhlOrV] c|g|l|t|x CABFILE\n"
       "\n"
       "Options:\n"
       "\t-c COMPONENT  Only list/extract this component\n"
@@ -112,6 +122,8 @@ static void show_usage(const char* name)
       "\t-h            Show this help message\n"
       "\t-j            Junk paths (do not make directories)\n"
       "\t-L            Make file and directory names lowercase\n"
+      "\t-O            Use old compression\n"
+      "\t-r            Save raw data (do not decompress)\n"
       "\t-V            Print copyright and version information\n"
       "\n"
       "Commands:\n"
@@ -140,7 +152,7 @@ static bool handle_parameters(
 {
 	int c;
 
-	while ((c = getopt(argc, argv, "c:d:D:g:hjLnoV")) != -1)
+	while ((c = getopt(argc, argv, "c:d:D:g:hjLnoOrV")) != -1)
 	{
 		switch (c)
     {
@@ -176,6 +188,14 @@ static bool handle_parameters(
         overwrite = OVERWRITE_ALWAYS;
         break;
 
+      case 'O':
+        format = FORMAT_OLD;
+        break;
+        
+      case 'r':
+        format = FORMAT_RAW;
+        break;
+
       case 'v':
         verbose = true;
         break;
@@ -305,7 +325,18 @@ static bool extract_file(Unshield* unshield, const char* prefix, int index)
   }
 
   printf("  extracting: %s\n", filename);
-  success = unshield_file_save(unshield, index, filename);
+  switch (format)
+  {
+    case FORMAT_NEW:
+      success = unshield_file_save(unshield, index, filename);
+      break;
+    case FORMAT_OLD:
+      success = unshield_file_save_old(unshield, index, filename);
+      break;
+    case FORMAT_RAW:
+      success = unshield_file_save_raw(unshield, index, filename);
+      break;
+  }
 
   if (!success)
   {
@@ -417,7 +448,10 @@ static int list_files_helper(Unshield* unshield, const char* prefix, int first,
       valid_count++;
 
       if (prefix && prefix[0])
+      {
         strcpy(dirname, prefix);
+        strcat(dirname, "\\");
+      }
       else
         dirname[0] = '\0';
 
@@ -430,8 +464,8 @@ static int list_files_helper(Unshield* unshield, const char* prefix, int first,
           *p = '/';
 #endif
 
-      if (dirname[strlen(dirname)-1] != '/')
-        strcat(dirname, "/");
+      if (dirname[strlen(dirname)-1] != '\\')
+        strcat(dirname, "\\");
 
       printf(" %8" SIZE_FORMAT "  %s%s\n",
           unshield_file_size(unshield, i),
